/************************************************************************
 *   IRC - Internet Relay Chat, modules/database/ccr.c
 *   This file is copyright (C) 2001 Andrew Suffield
 *                                    <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "config.h"
#include "database.h"
#include "db_internals.h"
#include "client.h"
#include "send.h"
#include "umodes.h"
#include "irc_string.h"
#include "s_log.h"
#include "ccr.h"
#include "numeric.h"
#include "ircd.h"

#include <db.h>
#include <stdlib.h>

/* Returns -1 on parse error */
static inline int
ccr_from_string(struct client_connect_record *ccr, char *p, unsigned int size)
{
  if (size < (sizeof(u_int16_t) + 3 * sizeof(time_t) + sizeof(u_int32_t)))
    return -1;
  ccr->priority = *((const u_int16_t *)p);
  p += sizeof(u_int16_t);
  ccr->when = *((const time_t *)p);
  p += sizeof(time_t);
  ccr->until = *((const time_t *)p);
  p += sizeof(time_t);
  ccr->time_deleted = *((const time_t *)p);
  p += sizeof(time_t);
  ccr->flags = *((const u_int32_t *)p);
  p += sizeof(u_int32_t);
  ccr->class = *((const u_int16_t *)p);
  p += sizeof(u_int16_t);

  p += extract_pascal_string(&ccr->host, p);
  p += extract_pascal_string(&ccr->placed_by, p);
  p += extract_pascal_string(&ccr->reason, p);
  p += extract_pascal_string(&ccr->deleted_by, p);
  p += extract_pascal_string(&ccr->spoof_host, p);
  p += extract_pascal_string(&ccr->passwd, p);

  return 0;
}

/* Returns the length of the buffer */
static inline int
string_from_ccr(char *buf, const struct client_connect_record *ccr)
{
  /* OK, build the data buffer */
  char *p = buf;

  *((u_int16_t *)p) = ccr->priority;
  p += sizeof(u_int16_t);
  *((time_t *)p) = ccr->when;
  p += sizeof(time_t);
  *((time_t *)p) = ccr->until;
  p += sizeof(time_t);
  *((time_t *)p) = ccr->time_deleted;
  p += sizeof(time_t);
  *((u_int32_t *)p) = ccr->flags;
  p += sizeof(u_int32_t);
  *((u_int16_t *)p) = ccr->class;
  p += sizeof(u_int16_t);

  p += add_pascal_string(p, ccr->host);
  p += add_pascal_string(p, ccr->placed_by);
  p += add_pascal_string(p, ccr->reason);
  p += add_pascal_string(p, ccr->deleted_by);
  p += add_pascal_string(p, ccr->spoof_host);
  p += add_pascal_string(p, ccr->passwd);

  return p - buf;
}

static inline int
string_from_idr(char *buf, struct ip_drop_record *idr)
{
  char *p = buf;

  *((unsigned long *)p) = idr->ip;
  p += sizeof(unsigned long);
  *((unsigned long *)p) = idr->ip_mask;
  p += sizeof(unsigned long);
  *((time_t *)p) = idr->placed_at;
  p += sizeof(time_t);
  *((time_t *)p) = idr->until;
  p += sizeof(time_t);
  *((time_t *)p) = idr->deleted_at;
  p += sizeof(time_t);

  p += add_pascal_string(p, idr->placed_by);
  p += add_pascal_string(p, idr->deleted_by);

  return p - buf;
}

static inline int
idr_from_string(struct ip_drop_record *idr, char *p, unsigned int size)
{
  if (size < sizeof(unsigned long) * 2 + sizeof(time_t) * 3)
    return -1;

  idr->ip = *((unsigned long *)p);
  p += sizeof(unsigned long);
  idr->ip_mask = *((unsigned long *)p);
  p += sizeof(unsigned long);

  idr->placed_at = *((time_t *)p);
  p += sizeof(time_t);
  idr->until = *((time_t *)p);
  p += sizeof(time_t);
  idr->deleted_at = *((time_t *)p);
  p += sizeof(time_t);

  p += extract_pascal_string(&idr->placed_by, p);
  p += extract_pascal_string(&idr->deleted_by, p);

  return 0;
}

/* Returns 0 if found, 1 if not found, -1 to fail silently (already broadcast error)
 *  and a db error code otherwise
 */
static inline int
hunt_host_entry(DBC *dbc, char *keyval, const char *host, struct client_connect_record *ccr, int exact)
{
  /* static because ccr_from_string()'s output is only valid while the string exists */
  static DBT key, data;
  int rc;

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  key.data = keyval;
  key.size = strlen(keyval) + 1;

  for (;;)
    switch(rc = dbc->c_get(dbc, &key, &data, DB_SET))
      {
      case 0:
	if (ccr_from_string(ccr, data.data, data.size))
	  {
	    broadcast_alert("ccr_from_string() failed, ccr database corrupted");
	    return -1;
	  }
	
	/* Loop until found */
	while (ccr->time_deleted || !(ccr->host && (exact ? !irccmp(ccr->host, host) : match(ccr->host, host))))
	  switch(rc = dbc->c_get(dbc, &key, &data, DB_NEXT_DUP))
	    {
	    case 0:
	      if (ccr_from_string(ccr, data.data, data.size))
		{
		  broadcast_alert("ccr_from_string() failed, ccr database corrupted");
		  return -1;
		}
	      break;
	    case DB_LOCK_DEADLOCK:
	      continue;
	    default:
	      return rc;
	    }
	/* If we drop to here, we found it */
	return 0;
      case DB_LOCK_DEADLOCK:
	continue;
      default:
	return rc;
      }

  /* not reached */
  abort();
}

/* Abstraction is always handy ;) */
static inline struct client_connect_record *
find_prevalent_ccr(DB *db, const char *host, struct client_connect_record *ccr)
{
  int rc;
  char buf[HOSTLEN + 2];
  char *clip_host;
  int len = strlen(host);
  DBC *dbc;

  /* Safety */
  if (!len)
    return NULL;

  strncpy_irc(buf, host, len);
  /* OK, we need the trailing . if it's not already there */
  if (buf[len - 1] != '.')
    {
      buf[len++] = '.';
      buf[len] = '\0';
    }
  clip_host = buf;

  if ((rc = db->cursor(db, NULL, &dbc, 0)))
    return broadcast_fail_null("find_prevalent_ccr: db->cursor()", rc);

  /* OK, here's how it works. A ccr for mask abc.def.ghi.jkl will
   *  be stored with key abc.def.ghi.jkl. One for *.def.ghi.jkl will
   *  be stored with key def.ghi.jkl., and so on.
   * Notice the trailing period in the keys
   * The key . is where the old "unsortable" list is to be found
   * The key is basically the domain chunk up to the first segment which has
   *  a wildcard in it, but not including this segment.
   *
   * We search for a given host foo.bar.baz.com, by first looking up the
   *  key foo.bar.baz.com., then bar.baz.com., then baz.com., then com.,
   *  then lastly . itself.
   */

  while (clip_host && *clip_host)
    {
      switch ((rc = hunt_host_entry(dbc, clip_host, host, ccr, 0)))
	{
	case 0:
	  /* Bingo. Return this */
	  if ((rc = dbc->c_close(dbc)))
	    return broadcast_fail_null("find_prevalent_ccr: dbc->c_close", rc); /* or not... */
	  return ccr;
	case DB_NOTFOUND:
	  /* Not this time, try again with a shorter key */
	  break;
	default:
	  /* Eep, something wicked happened */
	  if (rc != -1) /* Special case in hunt_host_entry() */
	    broadcast_fail("hunt_host_entry", rc);
	  if ((rc = dbc->c_close(dbc)))
	    broadcast_fail("find_prevalent_ccr: dbc->c_close", rc);
	  return NULL;
	}

      /* Clip off the first chunk and try again */
      clip_host = strchr(clip_host + 1, '.');
      /* Skip over the . unless it's the last one */
      if (clip_host && (*(clip_host + 1) != '\0'))
	clip_host++;
    }

  if ((rc = dbc->c_close(dbc)))
    broadcast_fail("find_prevalent_ccr: dbc->c_close", rc);
  return NULL;
}

struct client_connect_record *
find_prevalent_allow_ccr(const char *host)
{
  static struct client_connect_record ccr;
  memset(&ccr, 0, sizeof(ccr));
  return find_prevalent_ccr(allow_ccr_db, host, &ccr);
}

struct client_connect_record *
find_prevalent_deny_ccr(const char *host)
{
  static struct client_connect_record ccr;
  memset(&ccr, 0, sizeof(ccr));
  return find_prevalent_ccr(deny_ccr_db, host, &ccr);
}

static inline char *
get_key(const char *host)
{
  static char buf[HOSTLEN + 2];
  const char *p, *q;

  /* Figure out what the key will be. It's all the domain parts on the RHS
   *  I can grab which don't contain wilds
   */
  if ((p = strrchr(host, '*')))
    p = host;

  /* And skip over any ?s as well... */
  if (!(q = strrchr(p, '?')))
    q = p;

  /* If we're still at the start, don't chop the first segment
   * Unless the first character is a wild, of course
   */
  if ((q == host) && (*q != '*') && (*q != '?'))
    {
      strcpy(buf, host);
    }
  else
    {
      /* OK, after the last wild, grab the first . */
      if (!(p = strchr(q, '.')))
	{
	  /* There isn't a . after this, so our key is just '.' */
	  buf[0] = '.';
	  buf[1] = '\0';
	}
      else
	{
	  /* If this is not the last character, skip the .
	   *  This turns .b.c into b.c, but leaves . alone
	   */
	  if (*(p + 1) != '\0')
	    p++;
	  
	  /* Whatever is left now is our key */
	  strcpy(buf, p);
	}
    }
  if (buf[strlen(buf) - 1] != '.')
    strcat(buf, ".");

  return buf;
}

/* This searches the duplicate list beginning with whatever key has in it
 *  for entries matching ccr. If a perfect (priority and host) match is found,
 *  then return 0 with the cursor pointing to that record, so that it can be
 *  updated.
 * If a partial (host but not priority) match is found, that record is deleted
 *  and DB_NOTFOUNT is returned, since it has to be added in a different location
 */
static inline int
find_ccr(DBC *dbc, DBT *key, DBT *data, struct client_connect_record *ccr)
{
  int rc;
  struct client_connect_record this_ccr;

  switch(rc = dbc->c_get(dbc, key, data, DB_SET))
    {
    case 0:
      break;
    case DB_NOTFOUND:
      /* Weird. Ah well */
      return DB_NOTFOUND;
    case DB_LOCK_DEADLOCK:
      /* Grr. */
      return DB_LOCK_DEADLOCK;
    default:
      return broadcast_fail("find_ccr: dbc->c_get(..., DB_SET)", rc);
    }

  if (ccr_from_string(&this_ccr, data->data, data->size))
    {
      broadcast_alert("ccr_from_string() failed, ccr database corrupted");
      return -1;
    }

  while(irccmp(ccr->host, this_ccr.host))
    {
      switch(rc = dbc->c_get(dbc, key, data, DB_NEXT_DUP))
	{
	case 0:
	  break;
	case DB_LOCK_DEADLOCK:
	case DB_NOTFOUND:
	  /* Pass these two back up */
	  return rc;
	default:
	  return broadcast_fail("find_ccr: dbc->c_get(..., DB_NEXT_DUP)", rc);
	}

      if (ccr_from_string(&this_ccr, data->data, data->size))
	{
	  broadcast_alert("ccr_from_string() failed, ccr database corrupted");
	  return -1;
	}
    }

  /* OK, we found a match */
  if (ccr->priority == this_ccr.priority)
    /* Perfect match */
    return 0;

  /* Delete it */
  switch(rc = dbc->c_del(dbc, 0))
    {
    case 0:
      return DB_NOTFOUND;
    case DB_LOCK_DEADLOCK:
      return DB_LOCK_DEADLOCK;
    default:
      return broadcast_fail("find_ccr: dbc->c_del", rc);
    }

  return DB_NOTFOUND;
}

static inline int
try_store_ccr(DB *db, DB_TXN *tid, struct client_connect_record *ccr)
{
  char buf[65535];
  int rc;
  char *keyval;
  DBT key, data, ckey, cdata;
  DBC *dbc;

  if (!ccr->host)
    {
      sendto_ops_flag(UMODE_DEBUG, "WTF: ccr->host is NULL in try_store_ccr()");
      return -1;
    }
  keyval = get_key(ccr->host);

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  memset(&ckey, 0, sizeof(key));
  memset(&cdata, 0, sizeof(cdata));
  ckey.data = key.data = keyval;
  ckey.size = key.size = strlen(keyval) + 1;
  data.size = string_from_ccr(buf, ccr);
  data.data = buf;

  /* First, assume there is no other key in there, just try and shove it in */
  switch(rc = db->put(db, tid, &key, &data, DB_NOOVERWRITE))
    {
    case 0:
      /* Success */
      return 0;
    case DB_LOCK_DEADLOCK:
      return DB_LOCK_DEADLOCK;
    case DB_KEYEXIST:
      /* Hrmph, need to go do a cursor search thing */
      if ((rc = db->cursor(db, tid, &dbc, 0)))
	return broadcast_fail("try_store_ccr: db->cursor()", rc);
      switch(find_ccr(dbc, &ckey, &cdata, ccr))
	{
	case 0:
	  /* Replace the current record with this new one */
	  switch(rc = dbc->c_put(dbc, &key, &data, DB_CURRENT))
	    {
	    case 0:
	      if ((rc = dbc->c_close(dbc)))
		return broadcast_fail("try_store_ccr: dbc->c_close", rc);
	      else
		return 0;
	    case DB_LOCK_DEADLOCK:
	      if ((rc = dbc->c_close(dbc)))
		return broadcast_fail("try_store_ccr: dbc->c_close", rc);
	      else
		return DB_LOCK_DEADLOCK;
	    default:
	      broadcast_fail("try_store_ccr: dbc->c_put(..., DB_CURRENT)", rc);
	      if ((rc = dbc->c_close(dbc)))
		broadcast_fail("try_store_ccr: dbc->c_close", rc);
	      return -1;
	    }
	case DB_LOCK_DEADLOCK:
	  if ((rc = dbc->c_close(dbc)))
	    return broadcast_fail("try_store_ccr: dbc->c_close", rc);
	  else
	    return DB_LOCK_DEADLOCK;
	case DB_NOTFOUND:
	  if ((rc = dbc->c_close(dbc)))
	    return broadcast_fail("try_store_ccr: dbc->c_close", rc);
	  /* Ok, this is a new record, we really _can_ add it */
	  /* (or find_ccr deleted the old one for us) */
	  switch(rc = db->put(db, tid, &key, &data, DB_NOOVERWRITE))
	    {
	    case 0:
	      return 0;
	    case DB_LOCK_DEADLOCK:
	      return DB_LOCK_DEADLOCK;
	    default:
	      return broadcast_fail("try_store_ccr: db->put()", rc);
	    }
	default:
	  if ((rc = dbc->c_close(dbc)))
	    broadcast_fail("try_store_ccr: dbc->c_close", rc);
	  return -1;
	}
    default:
      return broadcast_fail("try_store_ccr: db->put(..., DB_NOOVERWRITE)", rc);
    }

  return 0;
}

static inline int
store_ccr(DB *db, struct client_connect_record *ccr)
{
  DB_TXN *tid;
  int rc, i;

  if ((rc = txn_begin(dbenv, NULL, &tid, 0)))
    return broadcast_fail("txn_begin", rc);

  /* Call a subfunction so that we can handle all the transaction stuff once here */
  for(i = 0; i < RETRY_COUNT; i++)
    {
      switch(try_store_ccr(db, tid, ccr))
	{
	case 0:
	  if ((rc = txn_commit(tid, 0)))
	    return broadcast_fail("store_ccr: txn_commit", rc);
	  else
	    return 0;
	case DB_LOCK_DEADLOCK:
	  /* Drat. Try again if we can */
	  if ((rc = txn_abort(tid)))
	    return broadcast_fail("store_ccr: txn_abort", rc);
	  else
	    continue;
	default:
	  if ((rc = txn_abort(tid)))
	    return broadcast_fail("store_ccr: txn_abort", rc);
	  else
	    return -1;
	}
    }

  broadcast_alert("Too many transaction failures in store_ccr()");
  return -1;
}

int
store_allow_ccr(struct client_connect_record *ccr)
{
  return store_ccr(allow_ccr_db, ccr);
}

int 
store_deny_ccr(struct client_connect_record *ccr)
{
  return store_ccr(deny_ccr_db, ccr);
}

static inline int
search_ccr(DB *db, struct Client *sptr, char *mask, int show_deleted)
{
  DBC *dbc;
  DBT key, data;
  struct client_connect_record ccr;
  int rc;

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  if ((rc = db->cursor(db, NULL, &dbc, 0)))
    return broadcast_fail("search_ccr: db->cursor()", rc);

  for(;;)
    switch(rc = dbc->c_get(dbc, &key, &data, DB_NEXT))
      {
      case 0:
	if (ccr_from_string(&ccr, data.data, data.size))
	  {
	    broadcast_alert("ccr_from_string() failed, ccr database corrupted");
	    return -1;
	  }
	if ((show_deleted && !ccr.time_deleted)
	    || (!show_deleted && ccr.time_deleted))
	  continue;

	/* Safety check */
	if (!ccr.host)
	  continue;

	if (!match(mask, ccr.host))
	  continue;

	sendto_one(sptr, form_str(RPL_CCONNECT), me.name, sptr->name,
		   ccr.host, ccr.priority, ccr.placed_by, ccr.when,
		   ccr.until, ",",
		   ccr.deleted_by ? ccr.deleted_by : "-",
		   ccr.time_deleted,
		   ccr.reason ? ccr.reason : "");

	break;
      case DB_LOCK_DEADLOCK:
	/* OH COME ON! */
	continue;
      case DB_NOTFOUND:
	sendto_one(sptr, form_str(RPL_CCONNECT_END), me.name, sptr->name);
	if ((rc = dbc->c_close(dbc)))
	  return broadcast_fail("search_ccr: dbc->c_close", rc);
	return 0;
      default:
	broadcast_fail("search_ccr: dbc->c_get(..., DB_NEXT)", rc);
	if ((rc = dbc->c_close(dbc)))
	  broadcast_fail("search_ccr: dbc->c_close", rc);
	return -1;
      }
}

int
search_allow_ccr(struct Client *sptr, char *mask, int show_deleted)
{
  return search_ccr(allow_ccr_db, sptr, mask, show_deleted);
}

int
search_deny_ccr(struct Client *sptr, char *mask, int show_deleted)
{
  return search_ccr(deny_ccr_db, sptr, mask, show_deleted);
}

int
ccr_dup_compare(DB *db, const DBT *a, const DBT *b)
{
  u_int16_t priority_a, priority_b;
  if ((a->size < 2) || (b->size < 2))
    return 0;
  priority_a = *((const u_int16_t *)a->data);
  priority_b = *((const u_int16_t *)b->data);
  return (priority_a - priority_b);
}

/* FIXME: This should use a transaction */
static inline int
delete_ccr(DB *db, struct Client *sptr, char *host)
{
  int rc;
  char buf[65535];
  char *keyval;
  struct client_connect_record ccr;
  DBC *dbc;
  DBT key, data;

  keyval = get_key(host);

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  if ((rc = db->cursor(db, NULL, &dbc, 0)))
    return broadcast_fail("delete_ccr: db->cursor()", rc);

  switch ((rc = hunt_host_entry(dbc, keyval, host, &ccr, 1)))
    {
    case 0:
      /* Bingo. Kill it */
      ccr.deleted_by = sptr->name;
      ccr.time_deleted = CurrentTime;
      key.data = keyval;
      key.size = strlen(keyval) + 1;
      data.size = string_from_ccr(buf, &ccr);
      data.data = buf;
      for(;;)
	switch(rc = dbc->c_put(dbc, &key, &data, DB_CURRENT))
	  {
	  case 0:
	    if ((rc = dbc->c_close(dbc)))
	      return broadcast_fail("delete_ccr: dbc->c_close", rc);
	    else
	      return 0;
	  case DB_LOCK_DEADLOCK:
	    continue;
	  default:
	    broadcast_fail("delete_ccr: dbc->c_put(..., DB_CURRENT)", rc);
	    if ((rc = dbc->c_close(dbc)))
	      broadcast_fail("delete_ccr: dbc->c_close", rc);
	    return -1;
	  }
      abort();
    case DB_NOTFOUND:
      /* Ignoreth */
      break;
    default:
      /* Eep, something wicked happened */
      if (rc != -1) /* Special case in hunt_host_entry() */
	broadcast_fail("hunt_host_entry", rc);
      if ((rc = dbc->c_close(dbc)))
	broadcast_fail("delete_ccr: dbc->c_close", rc);
      return -1;
    }

  if ((rc = dbc->c_close(dbc)))
    broadcast_fail("delete_ccr: dbc->c_close", rc);
  return 1;
}

int
delete_allow_ccr(struct Client *sptr, char *host)
{
  return delete_ccr(allow_ccr_db, sptr, host);
}

int
delete_deny_ccr(struct Client *sptr, char *host)
{
  return delete_ccr(deny_ccr_db, sptr, host);
}

int
add_ip_drop(struct ip_drop_record *idr)
{
  DBT key, data;
  char buf[65535];
  DB_TXN *tid;
  int rc;

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  key.data = &idr->ip;
  key.size = sizeof(idr->ip);
  data.size = string_from_idr(buf, idr);
  data.data = buf;

  for(;;)
    {
      if ((rc = txn_begin(dbenv, NULL, &tid, 0)))
	return broadcast_fail("txn_begin", rc);

      switch(rc = ip_drop_db->put(ip_drop_db, tid, &key, &data, 0))
	{
	case 0:
	  if ((rc = txn_commit(tid, 0)))
	    return broadcast_fail("add_ip_drop: txn_commit", rc);
	  else
	    {
	      /* OK, we have to copy it to a heap structure, this one is
	       *  getting kept in memory
	       */
	      struct ip_drop_record *pidr;
	  
	      pidr = MyMalloc(sizeof(struct ip_drop_record));
	      memcpy(pidr, idr, sizeof(struct ip_drop_record));
	      if (idr->placed_by)
		DupString(pidr->placed_by, idr->placed_by);
	      if (idr->deleted_by)
		DupString(pidr->deleted_by, idr->deleted_by);
	      if (idr->reason)
		DupString(pidr->reason, idr->reason);

	      /* Shove it in the tree thing */
	      if (pidr->exception)
		add_ip_drop_exception_entry(pidr);
	      else
		add_ip_drop_entry(pidr);

	      return 0;
	    }
	case DB_LOCK_DEADLOCK:
	  /* Drat. Try again if we can */
	  if ((rc = txn_abort(tid)))
	    return broadcast_fail("add_ip_drop: txn_abort", rc);
	  else
	    continue;
	default:
	  broadcast_fail("add_ip_drop: ip_drop_db->put", rc);
	  if ((rc = txn_abort(tid)))
	    return broadcast_fail("add_ip_drop: txn_abort", rc);
	  else
	    return -1;
	}
      abort();
    }

  abort();
}

int
del_ip_drop(struct ip_drop_record *idr)
{
  DBT key;
  DB_TXN *tid;
  int rc;

  memset(&key, 0, sizeof(key));

  key.data = &idr->ip;
  key.size = sizeof(idr->ip);

  for(;;)
    {
      if ((rc = txn_begin(dbenv, NULL, &tid, 0)))
	return broadcast_fail("txn_begin", rc);

      switch(rc = ip_drop_db->del(ip_drop_db, tid, &key, 0))
	{
	case 0:
	  if ((rc = txn_commit(tid, 0)))
	    return broadcast_fail("del_ip_drop: txn_commit", rc);
	  else
	    {
	      /* Need to reload them, can't currently remove a single entry */
	      load_ip_drops();
	      return 0;
	    }
	case DB_LOCK_DEADLOCK:
	  /* Drat. Try again if we can */
	  if ((rc = txn_abort(tid)))
	    return broadcast_fail("del_ip_drop: txn_abort", rc);
	  else
	    continue;
	default:
	  broadcast_fail("del_ip_drop: ip_drop_db->put", rc);
	  if ((rc = txn_abort(tid)))
	    return broadcast_fail("del_ip_drop: txn_abort", rc);
	  else
	    return -1;
	}
      abort();
    }

  abort();
}

int
load_ip_drops()
{
  DBC *dbc;
  DBT key, data;
  struct ip_drop_record idr;
  int rc;

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  zap_ip_drops();

  if ((rc = ip_drop_db->cursor(ip_drop_db, NULL, &dbc, 0)))
    return broadcast_fail("load_ip_drops: ip_drop_db->cursor()", rc);

  for(;;)
    switch(rc = dbc->c_get(dbc, &key, &data, DB_NEXT))
      {
      case 0:
	{
	  if (idr_from_string(&idr, data.data, data.size))
	    {
	      broadcast_alert("idr_from_string() failed, ip drop database corrupted");
	      return -1;
	    }
	  if (idr.deleted_at)
	    continue;

	  {
	    /* OK, we have to copy it to a heap structure, this one is
	     *  getting kept in memory
	     */
	    struct ip_drop_record *pidr;
	  
	    pidr = MyMalloc(sizeof(struct ip_drop_record));
	    memcpy(pidr, &idr, sizeof(idr));
	    if (idr.placed_by)
	      DupString(pidr->placed_by, idr.placed_by);
	    if (idr.deleted_by)
	      DupString(pidr->deleted_by, idr.deleted_by);
	    if (idr.reason)
	      DupString(pidr->reason, idr.reason);

	    /* Shove it in the tree thing */
	    if (pidr->exception)
	      add_ip_drop_exception_entry(pidr);
	    else
	      add_ip_drop_entry(pidr);
	  }

	  break;
	}
      case DB_LOCK_DEADLOCK:
	/* OH COME ON! */
	continue;
      case DB_NOTFOUND:
	if ((rc = dbc->c_close(dbc)))
	  return broadcast_fail("load_ip_drops: dbc->c_close", rc);
	return 0;
      default:
	broadcast_fail("load_ip_drops: dbc->c_get(..., DB_NEXT)", rc);
	if ((rc = dbc->c_close(dbc)))
	  broadcast_fail("load_ip_drops: dbc->c_close", rc);
	return -1;
      }
  return 0;
}

int
search_ip_drops(struct Client *sptr, char *mask)
{
  return 0;
}

void
free_ip_drop(struct ip_drop_record *idr)
{
  if (!idr)
    return;

  MyFree(idr->placed_by);
  MyFree(idr->deleted_by);
  MyFree(idr->reason);
  MyFree(idr);
}
